from __future__ import annotations

from dataclasses import dataclass, fields
from typing import Type, Generator, get_type_hints, get_args, Dict

from google.cloud import datastore


class ClientMixin:
    _client = None

    @classmethod
    def _get_client(cls) -> datastore.Client:
        if not cls._CLIENT:
            cls._CLIENT = datastore.Client()
        return cls._CLIENT


@dataclass
class DoesNotExist(Exception):
    cls: Type[EmbeddedDocument]
    pk: str


@dataclass
class MultipleObjectsFound(Exception):
    cls: Type[EmbeddedDocument]
    filters: Dict


operators = {
    'eq': '==',
    'gt': '>',
    'gte': '>=',
    'lt': '<',
    'lte': '<=',
    'in': 'in',
}


def query_operator(key: str) -> tuple:
    parts = key.split('__')
    field = parts[0]
    if len(parts) == 1:
        operator = '=='
    else:
        try:
            operator = operators[parts[1]]
        except KeyError as e:
            raise Exception(f"Unsupported query operator {parts[1]}") from e
    return field, operator


@dataclass
class EmbeddedDocument:
    @classmethod
    def _fields(cls):
        resolved_hints = get_type_hints(cls)
        field_names = [field.name for field in fields(cls)]
        return {
            name: resolved_hints[name]
            for name in field_names
            if not name.startswith('_')
        }

    @classmethod
    def deserialize(cls, **kwargs) -> EmbeddedDocument:
        return cls._from_dict(**kwargs)

    @classmethod
    def _from_dict(cls, **kwargs) -> EmbeddedDocument:
        data = kwargs.copy()

        def _build(klass, value):
            if value is None:
                return value

            if issubclass(klass, EmbeddedDocument):
                return klass._from_dict(**value)
            return klass(value)

        parsed_data = {}
        for field_name, field_klass in cls._fields().items():
            try:
                raw_value = data[field_name]
            except KeyError:
                continue

            if getattr(field_klass, '_name', '') == 'List':
                inner_klass = get_args(field_klass)[0]  # TODO: test composite types
                item = [_build(klass=inner_klass, value=i) for i in raw_value]
            elif getattr(field_klass, '_name', '') == 'Dict':
                inner_klass_key, inner_klass_value = get_args(field_klass)
                item = {
                    _build(klass=inner_klass_key, value=k): _build(klass=inner_klass_value, value=v)
                    for k, v in raw_value.items()
                }
            else:
                item = _build(klass=field_klass, value=raw_value)
            parsed_data[field_name] = item

        return cls(**parsed_data)

    def serialize(self) -> Dict:
        return self._to_dict()

    def _to_dict(self) -> dict:
        # TODO handle custom dynamic fields
        def _unbuild(value):
            if value is None:
                return value
            if isinstance(value, EmbeddedDocument):
                return value._to_dict()
            return value

        data = {}
        for field, field_klass in self.__class__._fields().items():
            raw_value = getattr(self, field)

            if getattr(field_klass, '_name', '') == 'List':
                item = [_unbuild(value=i) for i in raw_value]
            elif getattr(field_klass, '_name', '') == 'Dict':
                item = {
                    _unbuild(value=k): _unbuild(value=v)
                    for k, v in raw_value.items()
                }
            else:
                item = _unbuild(value=raw_value)

            data[field] = item

        return data


@dataclass
class Document(ClientMixin, EmbeddedDocument):
    @property
    def pk(self) -> str:
        raise NotImplementedError()

    @classmethod
    def _get_key(cls, pk: str = None) -> datastore.Key:
        return cls._get_client().key(cls._kind(), pk)

    @property
    def _key(self):
        return self._get_key(pk=self.pk)

    @classmethod
    def _kind(cls) -> str:
        return cls.__name__

    @classmethod
    def list(cls, **kwargs) -> Generator[Document, None, None]:
        query = cls._get_client().query(kind=cls._kind())

        for key, value in kwargs.items():
            field, operator = query_operator(key=key)
            query.add_filter(field, operator, value)

        for entity in query.fetch():
            yield cls._from_dict(**cls.from_entity(entity=entity))

    @classmethod
    def get(cls, pk: str = None, **kwargs) -> Document:
        if pk is not None:
            entity = cls._get_client().get(key=cls._get_key(pk=pk))
            if entity:
                return cls._from_dict(**cls.from_entity(entity=entity))
        else:
            one_obj = None
            for obj in cls.list(**kwargs):
                if one_obj is not None:
                    raise MultipleObjectsFound(cls, filters=kwargs)
                one_obj = obj
            return one_obj
        raise DoesNotExist(cls, pk)

    def to_entity(self) -> datastore.Entity:
        if not self.pk:
            raise Exception()

        entity = datastore.Entity(key=self._key)
        entity.update(self._to_dict())
        return entity

    @classmethod
    def from_entity(cls, entity: datastore.Entity) -> Dict:
        return dict(entity.items())

    @classmethod
    def create(cls, **kwargs) -> Document:
        data = kwargs.copy()
        obj = cls.deserialize(**data)
        return obj.save()

    def save(self) -> Document:
        self._get_client().put(entity=self.to_entity())
        return self.get(pk=self.pk)

    @classmethod
    def update(cls, pk: str, **kwargs) -> Document:
        if kwargs:
            entity = cls._get_client().get(key=cls._get_key(pk=pk))
            # TODO: enable partial nested updates
            as_data = {
                key: value._to_dict() if isinstance(value, EmbeddedDocument) else value
                for key, value in kwargs.items()
            }
            entity.update(as_data)
            cls._get_client().put(entity=entity)
        return cls.get(pk=pk)

    @classmethod
    def delete(cls, pk: str) -> None:
        cls._get_client().delete(key=cls._get_key(pk=pk))



# More Information: https://cloud.google.com/iam/docs/
from typing import Dict, Any, Generator

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI, AccountManagerMixin, PolicyType, DiscoveryMixin

AccountType = Dict[str, Any]


class IAM(AccountManagerMixin, DiscoveryMixin, GoogleCloudPilotAPI):
    def __init__(self, **kwargs):
        super().__init__(
            serviceName='iam',
            version='v1',
            cache_discovery=False,
            **kwargs,
        )

    def _parent_path(self, project_id: str = None) -> str:
        return f'projects/{project_id or self.project_id}'

    def _service_account_path(self, email: str, project_id: str = None) -> str:
        parent_path = self._parent_path(project_id=project_id)
        return f'{parent_path}/serviceAccounts/{email}'

    def _build_service_account_email(self, name: str, project_id: str = None) -> str:
        return f'{name}@{project_id or self.project_id}.iam.gserviceaccount.com'

    async def get_service_account(self, name: str, project_id: str = None) -> AccountType:
        account_path = self._service_account_path(
            email=self._build_service_account_email(name=name, project_id=project_id),
            project_id=project_id,
        )

        return self._execute(
            method=self.client.projects().serviceAccounts().get,
            name=account_path,
        )

    async def create_service_account(
            self,
            name: str,
            display_name: str,
            project_id: str = None,
            exists_ok: bool = True,
    ) -> AccountType:
        try:
            body = {
                'accountId': name,
                'serviceAccount': {
                    'displayName': display_name
                }
            }
            service_account = self._execute(
                method=self.client.projects().serviceAccounts().create,
                name=self._parent_path(project_id=project_id),
                body=body,
            )
        except exceptions.AlreadyExists:
            if not exists_ok:
                raise
            service_account = await self.get_service_account(name=name, project_id=project_id)
        return service_account

    async def list_service_accounts(self, project_id: str = None) -> Generator[AccountType, None, None]:
        params = dict(
            name=self._parent_path(project_id=project_id),
        )
        pagination = self._paginate(
            method=self.client.projects().serviceAccounts().list,
            result_key='accounts',
            params=params,
        )
        for item in pagination:
            yield item

    def get_policy(self, email: str, project_id: str = None) -> PolicyType:
        resource = self._service_account_path(email=email, project_id=project_id)
        return self._execute(
            method=self.client.projects().serviceAccounts().getIamPolicy,
            resource=resource,
        )

    def as_member(self, email: str) -> str:
        is_service_account = email.endswith('.gserviceaccount.com')
        prefix = 'serviceAccount' if is_service_account else 'member'
        return f'{prefix}:{email}'

    async def bind_member(self, target_email: str, member_email: str, role: str, project_id=None) -> PolicyType:
        policy = self.get_policy(email=target_email, project_id=project_id)
        changed_policy = self.bind_email_to_policy(email=member_email, role=role, policy=policy)
        return self.set_policy(email=target_email, policy=changed_policy, project_id=project_id)

    async def remove_member(
            self,
            target_email: str,
            member_email: str,
            role: str,
            project_id: str = None,
    ) -> PolicyType:
        policy = self.get_policy(email=target_email, project_id=project_id)
        changed_policy = self.unbind_email_from_policy(email=member_email, role=role, policy=policy)
        return self.set_policy(email=target_email, policy=changed_policy, project_id=project_id)

    def set_policy(self, email: str, policy: PolicyType, project_id: str = None) -> PolicyType:
        resource = self._service_account_path(email=email, project_id=project_id)
        return self._execute(
            method=self.client.projects().serviceAccounts().setIamPolicy,
            resource=resource,
            body={'policy': policy, 'updateMask': 'bindings'},
        )

    def get_compute_service_account(self, project_number: str = None) -> str:
        number = project_number or self._get_project_number(project_id=self.project_id)
        return f'{number}-compute@developer.gserviceaccount.com'

    def get_cloud_build_service_account(self, project_number: str = None) -> str:
        number = project_number or self._get_project_number(project_id=self.project_id)
        return f'{number}@cloudbuild.gserviceaccount.com'

    def get_cloud_run_service_account(self, project_number=None) -> str:
        number = project_number or self._get_project_number(project_id=self.project_id)
        return f'service-{number}@serverless-robot-prod.iam.gserviceaccount.com'



from contextlib import ExitStack
from typing import Any
from unittest.mock import Mock, patch


def patch_auth(credentials: Any = Mock(), project_id: str = 'potato-dev', location: str = 'moon-dark1'):
    managers = [
        patch('google.auth.default', return_value=(credentials, project_id)),
        patch('gcp_pilot.base.GoogleCloudPilotAPI._set_location', return_value=location)
    ]
    stack = ExitStack()
    for mgr in managers:
        stack.enter_context(mgr)
    return stack



import json

import requests

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI, DiscoveryMixin


class Text:
    @classmethod
    def build_mention(cls, member_id='all'):
        return f"<users/{member_id}>"

    @classmethod
    def build_link(cls, url, text):
        return f"<{url}|{text}>"


class Section:
    def __init__(self):
        self.header = None
        self.widgets = []

    def add_header(self, text):
        self.header = text

    def add_text(self, content, title='', footer='', click_url=None, icon=None, button=None):
        section = self._build_text_section(
            title=title,
            content=content,
            footer=footer,
            click_url=click_url,
            icon=icon,
            button=button,
        )
        self.widgets.append({'keyValue': section})

    def add_paragraph(self, text):
        section = self._build_paragraph(text=text)
        self.widgets.append({'textParagraph': section})

    def add_buttons(self, buttons):
        self.widgets.append({'buttons': buttons})

    def add_image(self, image_url, click_url=None):
        section = self._build_image(
            image_url=image_url,
            click_url=click_url,
        )
        self.widgets.append({'image': section})

    @classmethod
    def _build_paragraph(cls, text):
        return {'text': text}

    @classmethod
    def _build_image(cls, image_url, click_url=None):
        return {
            "imageUrl": image_url,
            "onClick": {"openLink": {"url": click_url or image_url}}
        }

    @classmethod
    def _build_text_section(
            cls,
            content,
            title='',
            footer='',
            click_url=None,
            icon=None,
            button=None,
    ):
        section = {
            "topLabel": title,
            "content": content,
            "contentMultiline": "true",
            "bottomLabel": footer,
        }

        if click_url:
            section["onClick"] = {
                "openLink": {
                    "url": click_url,
                }
            }

        if icon:
            section['icon'] = icon

        if button:
            section['button'] = button

        return section

    @classmethod
    def format_button(cls, url, text=None, image_url=None, icon=None):
        if text:
            button = {
                "textButton": {
                    "text": text,
                }
            }
        elif image_url:
            button = {
                "imageButton": {
                    "iconUrl": image_url,
                }
            }
        elif icon:
            button = {
                "imageButton": {
                    "icon": icon,
                }
            }
        else:
            raise exceptions.UnsupportedFormatException("A button must have a text, image or icon")

        button["onClick"] = {
            "openLink": {
                "url": url,
            }
        }
        return button

    @classmethod
    def format_color(cls, hex_color, text):
        return f"<font color=\"#{hex_color}\">{text}</font>"

    def as_data(self):
        data = {
            'widgets': self.widgets,
        }
        if self.header:
            data['header'] = self.header
        return data


class Card:
    def __init__(self):
        self.header = None
        self.sections = []

    def add_header(self, title, subtitle='', image_url=None, style='IMAGE'):
        self.header = self._build_header(
            title=title,
            subtitle=subtitle,
            image_url=image_url,
            style=style,
        )

    def add_section(self, section):
        self.sections.append(section)

    @classmethod
    def _build_header(cls, title, subtitle='', image_url=None, style='IMAGE'):
        return {
            "title": title,
            "subtitle": subtitle,
            "imageUrl": image_url,
            "imageStyle": style
        }

    def as_data(self):
        data = {'sections': [section.as_data() for section in self.sections]}
        if self.header:
            data['header'] = self.header
        return data


class ChatsHook:
    def __init__(self, hook_url):
        self.hook_url = hook_url

    def _post(self, body):
        response = requests.post(
            url=self.hook_url,
            headers={'Content-Type': 'application/json; charset=UTF-8'},
            data=json.dumps(body),
        )
        response.raise_for_status()
        return response.json()

    def send_text(self, text):
        body = {'text': text}
        return self._post(body=body)

    def send_card(self, card, additional_text=None):
        body = {
            'cards': [card.as_data()],
        }

        if additional_text:
            body['text'] = additional_text

        return self._post(body=body)


class ChatsBot(DiscoveryMixin, GoogleCloudPilotAPI):
    _scopes = ['https://www.googleapis.com/auth/chat.bot']

    def __init__(self, **kwargs):
        super().__init__(
            serviceName='chat',
            version='v1',
            cache_discovery=False,
            **kwargs,
        )

    def _room_path(self, room_id: str):
        prefix = 'spaces/'
        if not room_id.startswith(prefix):
            room_path = f'{prefix}{room_id}'
        else:
            room_path = room_id
        return room_path

    def _member_path(self, room_id: str, member_id: str):
        room_path = self._room_path(room_id=room_id)

        prefix = 'members/'
        if not member_id.startswith(prefix):
            member_path = f'{prefix}{member_id}'
        else:
            member_path = member_id

        return f'{room_path}/{member_path}'

    def get_room(self, room_id: str):
        return self._execute(
            method=self.client.spaces().get,
            name=self._room_path(room_id=room_id),
        )

    def get_rooms(self):
        yield from self._paginate(
            method=self.client.spaces().list,
            result_key='spaces',
        )

    def get_member(self, room_id: str, member_id: str):
        name = self._member_path(room_id=room_id, member_id=member_id)
        return self._execute(
            method=self.client.spaces().members().get,
            name=name,
        )

    def get_members(self, room_id: str):
        yield from self._paginate(
            method=self.client.spaces().members().list,
            result_key='memberships',
            params={'parent': self._room_path(room_id=room_id)}
        )

    def send_text(self, room_id: str, text: str):
        body = {'text': text}

        return self._execute(
            method=self.client.spaces().messages().create,
            parent=self._room_path(room_id=room_id),
            body=body,
        )

    def send_card(self, room_id: str, card: Card, additional_text: str = None):
        body = {
            'cards': [card.as_data()],
        }

        if additional_text:
            body['text'] = additional_text

        return self._execute(
            method=self.client.spaces().messages().create,
            parent=self._room_path(room_id=room_id),
            body=body,
        )



# https://googleapis.dev/python/pubsub/latest/index.html
from typing import Callable, Dict, Any, AsyncIterator

from google.api_core.exceptions import AlreadyExists, NotFound
from google.cloud import pubsub_v1
from google.pubsub_v1 import PushConfig, Subscription, types

from gcp_pilot.base import GoogleCloudPilotAPI


class CloudPublisher(GoogleCloudPilotAPI):
    _client_class = pubsub_v1.PublisherClient

    async def create_topic(self, topic_id: str, project_id: str = None, exists_ok: bool = True) -> types.Topic:
        topic_path = self.client.topic_path(
            project=project_id or self.project_id,
            topic=topic_id,
        )
        try:
            topic = self.client.create_topic(request={"name": topic_path})
        except AlreadyExists:
            if not exists_ok:
                raise
            topic = await self.get_topic(topic_id=topic_id, project_id=project_id)
        return topic

    async def get_topic(self, topic_id: str, project_id: str = None):
        pass

    async def publish(
            self,
            message: str,
            topic_id: str,
            project_id: str = None,
            attributes: Dict[str, Any] = None,
    ) -> types.PublishResponse:
        topic_path = self.client.topic_path(
            project=project_id or self.project_id,
            topic=topic_id,
        )
        try:
            future = self.client.publish(
                topic=topic_path,
                data=message.encode(),
                **(attributes or {}),
            )
            return future.result()
        except NotFound:
            await self.create_topic(
                topic_id=topic_id,
                project_id=project_id,
            )
            future = self.client.publish(
                topic=topic_path,
                data=message.encode(),
                **(attributes or {}),
            )
            return future.result()


class CloudSubscriber(GoogleCloudPilotAPI):
    _client_class = pubsub_v1.SubscriberClient

    async def list_subscriptions(self, project_id: str = None) -> AsyncIterator[Subscription]:
        all_subscriptions = self.client.list_subscriptions(
            project=f'projects/{project_id or self.project_id}',
        )
        for subscription in all_subscriptions:
            yield subscription

    async def get_subscription(self, subscription_id: str, project_id: str = None) -> Subscription:
        subscription_path = self.client.subscription_path(
            project=project_id or self.project_id,
            subscription=subscription_id,
        )

        return self.client.get_subscription(
            subscription=subscription_path,
        )

    async def delete_subscription(self, subscription_id: str, project_id: str = None) -> None:
        subscription_path = self.client.subscription_path(
            project=project_id or self.project_id,
            subscription=subscription_id,
        )

        return self.client.delete_subscription(
            subscription=subscription_path,
        )

    async def create_subscription(
            self,
            topic_id: str,
            subscription_id: str,
            project_id: str = None,
            exists_ok: bool = True,
            auto_create_topic: bool = True,
            push_to_url: str = None,
            use_oidc_auth: bool = False,
    ) -> Subscription:
        topic_path = self.client.topic_path(
            project=project_id or self.project_id,
            topic=topic_id,
        )
        subscription_path = self.client.subscription_path(
            project=project_id or self.project_id,
            subscription=subscription_id,
        )

        push_config = None
        if push_to_url:
            push_config = PushConfig(
                push_endpoint=push_to_url,
                **(self.oidc_token if use_oidc_auth else {}),
            )

        try:
            return self.client.create_subscription(
                name=subscription_path,
                topic=topic_path,
                push_config=push_config,
            )
        except NotFound:
            if not auto_create_topic:
                raise
            await CloudPublisher().create_topic(
                topic_id=topic_id,
                project_id=project_id,
                exists_ok=False,
            )
            return self.client.create_subscription(
                name=subscription_path,
                topic=topic_path,
                push_config=push_config,
            )
        except AlreadyExists:
            if not exists_ok:
                raise
            return await self.get_subscription(subscription_id=subscription_id, project_id=project_id)

    async def subscribe(self, topic_id: str, subscription_id: str, callback: Callable, project_id: str = None):
        await self.create_subscription(
            topic_id=topic_id,
            subscription_id=subscription_id,
            project_id=project_id,
        )

        subscription_path = self.client.subscription_path(
            project=project_id or self.project_id,
            subscription=subscription_id,
        )
        future = self.client.subscribe(
            subscription=subscription_path,
            callback=callback,
        )
        future.result()



class UnsupportedFormatException(Exception):
    pass


class ValidationError(Exception):
    pass


class UnboundException(Exception):
    pass


class NotFound(Exception):
    pass


class DeletedRecently(Exception):
    def __init__(self, resource, blocked_period='1 week'):
        message = f"{resource} was probably deleted recently. Cannot reuse name for {blocked_period}."
        super().__init__(message)


class BigQueryJobError(Exception):
    def __init__(self, job):
        message = ' | '.join([error['message'] for error in job.errors])
        super().__init__(message)


class AlreadyExists(Exception):
    pass


class AlreadyDeleted(Exception):
    pass


class NotAllowed(Exception):
    pass
